-- complain if script is sourced in psql, rather than via CREATE EXTENSION
\echo Use "CREATE EXTENSION mapi" to load this file. \quit

-- removes old functions from previous versions, if any.
DROP FUNCTION IF EXISTS MAPI_Version();
DROP FUNCTION IF EXISTS MAPI_MakeEllipse(geometry, geometry, float8, float8, integer);
DROP FUNCTION IF EXISTS MAPI_MakeHexagonalGrid(text, text, geometry, float8, boolean);
DROP FUNCTION IF EXISTS MAPI_Notice(IN text);
DROP FUNCTION IF EXISTS MAPI_Info(IN text);
DROP FUNCTION IF EXISTS MAPI_EstimateHalfwidth(geometry, int);
DROP FUNCTION IF EXISTS MAPI_Permute(text, text, text, text, text);
DROP FUNCTION IF EXISTS MAPI_OrderedConcat(text, text, text);
DROP FUNCTION IF EXISTS MAPI_EasyRunWithPermutations(text, text, text, text, float8, float8, float8, integer, float8, integer, integer, boolean, float8, float8);
DROP FUNCTION IF EXISTS MAPI_Polygonize(text, int, int, text);

-- TODO: clean before release
DROP FUNCTION IF EXISTS MAPI_BY_SGoF(float8[], float8);
DROP FUNCTION IF EXISTS MAPI_Bayesian_SGoF(float8[], float8);
DROP FUNCTION IF EXISTS MAPI_GridAuto(points_view_name text, grid_table_name text, my_schema_name text, beta float8);
DROP FUNCTION IF EXISTS MAPI_Significance_BenjaminiYekutieli(my_alpha float8, results_table text, my_schema_name text);
DROP FUNCTION IF EXISTS MAPI_Significance_BenjaminiYekutieli(results_table text, my_schema_name text, my_alpha float8, append boolean);
DROP FUNCTION IF EXISTS MAPI_Significance_Raw(my_threshold float8, results_table text, my_schema_name text);
DROP FUNCTION IF EXISTS MAPI_Significance_BY(results_table text, my_schema_name text, my_alpha float8, append boolean);
DROP FUNCTION IF EXISTS mapi_gridauto(text, text, text, double precision, boolean);
DROP FUNCTION IF EXISTS mapi_lbe(double precision[], double precision);
DROP FUNCTION IF EXISTS mapi_makeellipse(geometry, geometry, double precision, double precision, integer);
DROP FUNCTION IF EXISTS mapi_makehexagonalgrid(text, text, geometry, double precision, boolean);
DROP FUNCTION IF EXISTS mapi_polygonize(geometry);
DROP FUNCTION IF EXISTS mapi_rankinperms(double precision, double precision[]);
DROP FUNCTION IF EXISTS mapi_runauto(text, text, double precision, text, text, double precision, double precision, double precision, integer, boolean, double precision, double precision, double precision);
DROP FUNCTION IF EXISTS mapi_runongrid(text, text, text, text, text, double precision, double precision, double precision, integer, boolean, double precision, double precision);
DROP FUNCTION IF EXISTS mapi_significance_benjaminiyekutieli(double precision, text, text, boolean);
DROP FUNCTION IF EXISTS mapi_significance_lbe(double precision, text, text, boolean);
DROP FUNCTION IF EXISTS mapi_version();
DROP FUNCTION IF EXISTS MAPI_x_raster(results_table_name text, raster_table_name text, raster_band integer, my_schema_name text);
DROP FUNCTION IF EXISTS MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text, window_radius float8, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8);
DROP FUNCTION IF EXISTS MAPI_RunAuto(points_view_name text, relations_view_name text, beta float8, results_table_name text, my_schema_name text, window_radius float8, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8, alpha float8);
DROP FUNCTION IF EXISTS MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8);
DROP FUNCTION IF EXISTS MAPI_SimplifyTails(tails_table_name text, my_schema_name text, my_radius float8);
DROP AGGREGATE IF EXISTS MAPI_array_accum(anyarray);
DROP TYPE IF EXISTS weighted_value CASCADE;
--autodel: DROP FUNCTION IF EXISTS MAPI_stats(wv_arr weighted_value[], OUT nb int, OUT sumw float8, OUT avgwv float8, OUT varwv float8);

-- clean "addon" functions
DROP FUNCTION IF EXISTS MAPI_SGoF(float8[], varchar(250), float8, float8);
DROP FUNCTION IF EXISTS MAPI_Significance_SGoF(results_table text, my_schema_name text, my_method varchar(250), my_alpha float8, my_gamma float8, append boolean);
DROP FUNCTION IF EXISTS MAPI_LBE(float8[], float8);
DROP FUNCTION IF EXISTS MAPI_Significance_LBE(results_table text, my_schema_name text, my_alpha float8, append boolean);
DROP FUNCTION IF EXISTS MAPI_Kernel(results_table_name text, my_schema_name text, kernel varchar(250), bandwidth float8);


/**
SQL code for creation of functions involved in the MAPI process
	(c) S. Piry / INRA 2013-2016 unless otherwise mentioned.
	Licence: Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0)
	See https://creativecommons.org/licenses/by-nc-sa/3.0/deed.en_GB
**/


-- Designed by Sylvain Piry, CBGP - INRA, 2013-2016
-- Returns MAPI current version number and build date
CREATE OR REPLACE FUNCTION MAPI_Version()
  RETURNS text AS
$BODY$
BEGIN
  RETURN 'MAPI 1.0.1 $Revision: 136 $ $Date: 2016-04-01 13:08:45 +0200 (ven. 01 avril 2016) $';
END;
$BODY$ LANGUAGE plpgsql;
COMMENT ON FUNCTION MAPI_Version() IS 
'Designed by Sylvain Piry, CBGP - INRA, 2013-2016
Returns MAPI current version number and build date
';




-- Designed by Sylvain Piry, CBGP - INRA, 2015
-- Returns the rank in [0,1] of a given value in an array of values.
-- Counts the number of array members stictly below the value and divides by the array length.
DROP FUNCTION IF EXISTS MAPI_RankInDistribution(val float8, perms float8[]);
CREATE OR REPLACE FUNCTION MAPI_RankInDistribution(val float8, perms float8[])
  RETURNS float8 AS
$BODY$
DECLARE
  n integer;
  c integer;
BEGIN
  n := array_length(perms, 1);
  WITH q(p) AS (
	SELECT unnest(perms)
  )
  SELECT count(*)::integer
	FROM q
	WHERE p < val
  INTO c;
  RETURN (c::float8 / n::float8);
END;
$BODY$
  LANGUAGE plpgsql IMMUTABLE;
COMMENT ON FUNCTION MAPI_RankInDistribution(val float8, perms float8[]) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2015
Returns the rank in [0,1] of a given value in an array of values.
Counts the number of array members stictly below the value and divides by the array length.
';





-- Designed by Sylvain Piry & Astrid Cruaud, CBGP - INRA, 2013-2015
-- This function primarily builds a polygon approaching an ellipse, provided the two focus points, given the excentricity and the number of segments per quarter.
-- The geometry of the points have to be cartesian (no latitude / longitude).
-- An error circle may be defined around points. If the points are located at the same place a circle with given "err_radius" radius (in geometry projection units) is created instead.
-- Finally, the resulting polygon is the convex hull between the ellipse and a buffer of width "err_radius" around the segment joining the two points.
-- The number of segments per quarter allow to define the quality of the ellipsoidal polygon. Default is OK for current uses.
-- ARGS: point1 (geometry), point2 (geometry), eccentricity (double precision), error circle radius "err_radius" (double precision, optional), number of segments per quarter (integer, 8 by default).
DROP FUNCTION IF EXISTS MAPI_MakeEllipse(geometry(point), geometry(point), float8, float8, integer);
CREATE OR REPLACE FUNCTION MAPI_MakeEllipse(p1 geometry(point), p2 geometry(point), e float8, err_radius float8 DEFAULT NULL::float8, nb_seg integer DEFAULT 8)
  RETURNS geometry AS
$BODY$
DECLARE
  hp CONSTANT float8 := pi()/2.0;
  c1 CONSTANT geometry := ST_Buffer(ST_Point(0,0), 1, nb_seg);
  f float8; -- focal length
  a float8; -- major axis half-length
  b float8; -- minor axis half-length
  r float8; -- rotation angle
  sr float8; -- sin(r)
  cr float8; -- cos(r)
  s geometry; -- segment between the two points
  c geometry(point); -- centroid of the bipoint
  xc float8; -- x coordinate of the centroid
  yc float8; -- y coordinate of the centroid
  resu geometry; -- resulting ellipse more or less mixed with ovale, segment or point
BEGIN
  IF ((p2 IS NOT NULL) AND NOT ST_Equals(p1, p2)) THEN -- p1 and p2 are distinct points and both exists
    -- prepares variables
    s := ST_MakeLine(p1, p2);              -- segment between the two points
    IF e < 1.0 THEN -- true ellipse
      f := ST_Distance(p1, p2)/2.0;          -- focal length
      r := hp - ST_Azimuth(p1, p2);          -- rotation in radians 
      c := ST_Centroid(s);                   -- centroid of the bipoint
      xc := ST_X(c);
      yc := ST_Y(c);
      a := f / e;                            -- major axis half-length
      b := sqrt(a*a - f*f);                  -- minor axis half-length
      sr := sin(r);
      cr := cos(r);
      resu := ST_SetSRID(ST_Affine(c1, a*cr, -b*sr, 0, a*sr, b*cr, 0, 0, 0, 1, xc, yc, 0), ST_SRID(s));
    ELSE -- eccentricity = 1.0 --> ovale around error circles
      resu := ST_Buffer(s, 0); -- limited to the segment ; may be buffered below
    END IF;
  ELSE -- p1 and p2 are the same point or p2 is null. Assigns some variables anyway for further computations
    s := p1;
    resu := ST_Buffer(p1, 0);
  END IF;
  -- makes a buffer if err_radius provided
  IF err_radius IS NOT NULL THEN
    resu := ST_ConvexHull(ST_Collect(resu, ST_Buffer(s, err_radius)));
  END IF;
  -- returns the final geometry
  RETURN resu;
END;
$BODY$
  LANGUAGE plpgsql IMMUTABLE;
COMMENT ON FUNCTION MAPI_MakeEllipse(geometry, geometry, float8, float8, integer) IS 
'Designed by Sylvain Piry & Astrid Cruaud, CBGP - INRA, 2013-2015
This function primarily builds a polygon approaching an ellipse, provided the two focus points, 
    given the excentricity and the number of segments per quarter.
The geometry of the points have to be cartesian (no latitude / longitude).
An error circle may be defined around points. If the points are located at the same place a 
    circle with given "err_radius" radius (in geometry projection units) is created instead.
Finally, the resulting polygon is the convex hull between the ellipse and a buffer of width 
    "err_radius" around the segment joining the two points.
The number of segments per quarter allow to define the quality of the ellipsoidal polygon. 
    Default is OK for current uses.
Input parameters: 
  1/ point1 (geometry), 
  2/ point2 (geometry), 
  3/ eccentricity (double precision), 
  4/ error circle radius "err_radius" (double precision, optional), 
  5/ number of segments per quarter (integer, 8 by default).
Returns the polygon geometry.
';





-- Modified by Sylvain Piry CBGP - INRA, 2013, from: 
-- http://www.dimensionaledge.com/main/postgis/how-to-create-hexagonal-grids-in-postgis/
-- Builds hexagonal grid in "schemaname"."tablename" PostGis table covering a geomery set in the same SRID given the half-width of each hexagon.
-- ARGS : schemaname (text), tablename (text), the_zone (geometry), halfwidth (double precision), temporary_table (boolean, default false)
DROP FUNCTION IF EXISTS MAPI_GridHexagonal(text, text, geometry, float8, boolean);
CREATE OR REPLACE FUNCTION MAPI_GridHexagonal(schemaname text, tablename text, the_zone geometry, halfwidth float8, temporary_table boolean DEFAULT false)
  RETURNS integer AS
$BODY$
DECLARE
    srid text;
    bb box2d;
    x_value float8;
    y_value float8;
    x_count integer;
    y_count integer;
    y_offset float8;
    y_value_adj float8;
    sequencevar text;
    schemaname2 text;
    tablename2 text;
    fulltablename text;
    tbl_count integer := 0;
    cells_count integer;
    istmp text := '';
    final_comment text;
BEGIN
	final_comment := E'MAPI_GridHexagonal:\n';
	final_comment := final_comment || format(E'  schemaname=%s\n', coalesce(schemaname, 'NULL'));
	final_comment := final_comment || format(E'  tablename=%s\n', coalesce(tablename, 'NULL'));
	final_comment := final_comment || format(E'  the_zone=%s\n', coalesce(ST_AsText(the_zone), 'NULL'));
	final_comment := final_comment || format(E'  halfwidth=%s\n', coalesce(halfwidth::text, 'NULL'));
	final_comment := final_comment || format(E'  temporary_table=%s\n', coalesce(temporary_table::text, 'NULL'));
	
    IF the_zone IS NULL THEN RAISE 'No geometry provided'; END IF;
    IF halfwidth IS NULL THEN RAISE 'No halfwidth provided'; END IF;
    IF NOT halfwidth>0.0 THEN RAISE 'halfwidth must be >0'; END IF;
    srid := coalesce(ST_SRID(the_zone), 0)::text;
    schemaname2 := coalesce('"'||schemaname||'".', '');
    tablename2 := '"'||coalesce(tablename, 'hexgridtable')||'"';
    -- Create grid table
    IF temporary_table THEN 
	istmp := 'TEMPORARY'; 
	fulltablename := tablename2;
    ELSE
	istmp := ''; 
	fulltablename := schemaname2||tablename2;
    END IF;
    EXECUTE 'DROP TABLE IF EXISTS '||fulltablename;
    EXECUTE 'CREATE '||istmp||' TABLE '||fulltablename||' (gid serial PRIMARY KEY, x_count integer, x float8, y_count integer, y float8, geom geometry(polygon, '||srid||'));';
    -- load extents of the bounding box and values for the initial y parameters
    bb := ST_Extent(the_zone);
    RAISE INFO '*********** XMIN=%  XMAX=%  YMIN=%  YMAX=%   halfwidth=% **********', ST_XMin(bb), ST_XMax(bb), ST_YMin(bb), ST_YMax(bb), halfwidth;
    final_comment := final_comment || format(E'XMIN=%s  XMAX=%s  YMIN=%s  YMAX=%s\n', ST_XMin(bb), ST_XMax(bb), ST_YMin(bb), ST_YMax(bb));
    y_count := 1;
    y_value := st_ymin(bb);
    LOOP 
        -- for each y value, reset x to XMIN and subloop through the x values
        x_count := 1;
        x_value := st_xmin(bb);
        LOOP
            -- for every even numbered x_count, offset y by half the height of the hexagon
            y_offset := (halfwidth * sqrt(3) / 2.0) * (0.5 - (x_count % 2)::float8);
            -- add the offset to the y_value
            y_value_adj := y_value + y_offset;
            EXECUTE 'INSERT INTO '||fulltablename||' (x_count, x, y_count, y) VALUES('|| x_count ||', '|| x_value ||', ' || y_count ||', ' || y_value_adj ||')';
            tbl_count := tbl_count + 1;
            x_count := x_count + 1;
            x_value := x_value + halfwidth * 3.0 / 2.0;
            EXIT WHEN x_value > ST_XMax(bb);    
        END LOOP;
        -- after exiting the subloop, increment the y count and y value
        y_count := y_count + 1;
        y_value := y_value + halfwidth * sqrt(3);
        EXIT WHEN y_value > ST_YMax(bb);   
    END LOOP;

    -- With the table now populated with x y points, the last step is to create a hexagon geometry for each.
    EXECUTE 'UPDATE '||fulltablename||' SET geom = ST_Polygon(ST_MakeLine(ARRAY[
				ST_MakePoint(x-(0.5*'|| halfwidth ||'), y+(SQRT(3)*0.5*'|| halfwidth ||')), 
				ST_MakePoint(x-(1.0*'|| halfwidth ||'), y), 
				ST_MakePoint(x-(0.5*'|| halfwidth ||'), y-(SQRT(3)*0.5*'|| halfwidth ||')), 
				ST_MakePoint(x+(0.5*'|| halfwidth ||'), y-(SQRT(3)*0.5*'|| halfwidth ||')), 
				ST_MakePoint(x+(1.0*'|| halfwidth ||'), y), 
				ST_MakePoint(x+(0.5*'|| halfwidth ||'), y+(SQRT(3)*0.5*'|| halfwidth ||')), 
				ST_MakePoint(x-(0.5*'|| halfwidth ||'), y+(SQRT(3)*0.5*'|| halfwidth ||'))]
			),'||srid||')';
    -- Removes outliers
    EXECUTE 'DELETE FROM '||fulltablename||' WHERE NOT ST_Intersects(geom, ('''||ST_ConvexHull(the_zone)::text||''')::geometry)';
    -- Builds gist index on geometry
    EXECUTE 'CREATE INDEX "'||tablename||'_geom_gist" ON '||fulltablename||' USING gist (geom)';
    -- Returns number of cells
    EXECUTE 'SELECT count(*) FROM '||fulltablename INTO cells_count;
    RAISE INFO 'NUMBER OF CELLS: %', cells_count;
    final_comment := final_comment||format(E'Number of cells=%s\n', cells_count);
	EXECUTE 'COMMENT ON TABLE '||fulltablename||' IS '||quote_nullable(final_comment)||';';
    RETURN cells_count;
END;
$BODY$ 
LANGUAGE plpgsql VOLATILE;

COMMENT ON FUNCTION MAPI_GridHexagonal(text, text, geometry, float8, boolean) IS 
'Modified by Sylvain Piry CBGP - INRA, 2013, from: 
http://www.dimensionaledge.com/main/postgis/how-to-create-hexagonal-grids-in-postgis/
Builds hexagonal grid in "schemaname"."tablename" PostGis table covering a geomery set
    in the same SRID given the half-width of each hexagon.
Input parameters: 
  1/ schemaname (text),
  2/ tablename (text),
  3/ the_zone (geometry),
  4/ halfwidth (double precision).
Returns the number of cells.
';




-- Designed by Sylvain Piry, CBGP - INRA, 2013
-- Computes the half-width of hexagonal cells given a geometry and the number of hexagonal cells.
DROP FUNCTION IF EXISTS MAPI_EstimateHalfwidth(geometry, int);
CREATE OR REPLACE FUNCTION MAPI_EstimateHalfwidth(the_zone geometry, nb_cells int)
  RETURNS float8 AS
$BODY$
DECLARE
  zone_area float8;
  hex_area float8;
  hw float8;
BEGIN
  zone_area := ST_Area(ST_ConvexHull(the_zone));
  hex_area := zone_area / nb_cells;
  hw = sqrt(hex_area / 2.598076211);
  RAISE INFO 'ESTIMATED HALF-WIDTH: %', hw;
  return hw;
END;
$BODY$ LANGUAGE plpgsql VOLATILE;

COMMENT ON FUNCTION MAPI_EstimateHalfwidth(geometry, int) IS
'Designed by Sylvain Piry, CBGP - INRA, 2013
Computes the half-width of hexagonal cells given a geometry and 
	the approximative number of hexagonal cells expected.
Input parameters: 
  1/ the_zone (geometry),
  2/ nb_cells (integer)
Returns the value of the halfwidth.
';




-- Designed by Sylvain Piry, CBGP - INRA, 2015-2016
DROP FUNCTION IF EXISTS MAPI_GridAuto(points_view_name text, grid_table_name text, my_schema_name text, beta float8, full_extent boolean);
CREATE OR REPLACE FUNCTION MAPI_GridAuto(points_view_name text, grid_table_name text, my_schema_name text, beta float8 DEFAULT 0.25, full_extent boolean DEFAULT false)
	RETURNS integer AS
$BODY$
DECLARE 
	pts_tab text;
	srid integer;
	nb integer;
	z geometry;
	n float8;
	a float8;
	hw float8;
BEGIN 
	pts_tab := '"'||coalesce(my_schema_name||'"."','')||points_view_name||'"';
	EXECUTE 'SELECT ST_SRID(point_geom) FROM '||pts_tab||' LIMIT 1' INTO srid;
	IF (get_proj4_from_srid(srid) LIKE '%+proj=longlat%') THEN 
		RAISE 'Error in MAPI_GridAuto: Sampling points *must* be in cartesian coordinates.'; 
	END IF;
	
	IF (full_extent) THEN
		-- full extent of ellipses
		EXECUTE 'WITH 
				ch(g, n) AS (
					SELECT ST_ConvexHull(ST_Collect(point_geom)), count(DISTINCT point_geom)::float8
					FROM '||pts_tab||'
				),
				dp0(r, n) AS (
					SELECT ST_DumpPoints(g) 
					FROM ch
				),
				dp (p, g, n) AS (
					SELECT (r).*, n 
					FROM dp0
				),
				ell(e, n) AS (
					SELECT MAPI_MakeEllipse(p1.g, p2.g, 0.975, 0.001), p1.n
					FROM dp AS p1
					CROSS JOIN dp AS p2
				),
				q0(z, n) AS (
					SELECT ST_ConvexHull(ST_Collect(e)), n
					FROM ell
					GROUP BY n
				)
				SELECT z, n, ST_Area(z)
					FROM q0;' 
			INTO z, n, a;
	ELSE
		-- grid limited to points extent
		EXECUTE 'WITH
				q0(z, n) AS (
					SELECT ST_ConvexHull(ST_Collect(point_geom)), count(DISTINCT point_geom)::float8
					FROM '||pts_tab||'
				)
				SELECT z, n, ST_Area(z)
					FROM q0;' 
			INTO z, n, a;
	END IF;
	
	-- halfwidth (hw) is computed on the basis of the formula presented in user manual.
	hw := sqrt( ( (beta^2) / 2.598) * (a/n) );
	RAISE WARNING '*** AUTO-COMPUTED ***    halfwidth=%  (with: points data=% beta=%, area=%, nlocations=%)  to grid=%', hw, pts_tab, beta, a, n, '"'||coalesce(my_schema_name||'"."','')||grid_table_name||'"';
	
	-- calls: MAPI_GridHexagonal(schemaname text, tablename text, the_zone geometry, halfwidth float8, temporary_table boolean DEFAULT false)
	SELECT MAPI_GridHexagonal(my_schema_name, grid_table_name, z, hw, false) INTO nb;
	
	RETURN nb;
END;
$BODY$ 
	LANGUAGE plpgsql VOLATILE;
COMMENT ON FUNCTION MAPI_GridAuto(points_view_name text, grid_table_name text, my_schema_name text, beta float8, full_extent boolean) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2015-2016
Builds an grid of hexagonal cells on the basis of the samples locations.
By default (full_exent = false), the study area is the convex hull of sampling locations.
When full_extent is set to true, the grid will encompass the most external ellipses.
The beta parameter defines the density of cells within this area based on Shannon-Nyquist theorem.
beta=0.5 is suitable for regular samplings, beta=0.25 for random samplings.
Smaller beta values will increase the spatial resolution along with computation time.
Input parameters: 
  1/ samples_table name (text),
  2/ grid_table name (text),
  3/ schema name (text),
  4/ beta parameter (float),
  5/ full_extent (boolean, default FALSE).
Returns the number of cells in the grid.
';




-- Designed by Sylvain Piry, CBGP - INRA, 2013-2015
-- Aggregates contiguous cells in polygons for rank in permutations between min and max.
-- Parameters:
-- 1/ cells: geometry collection of polygons
-- Returns a set of rows (geom geometry(polygon))
DROP FUNCTION IF EXISTS MAPI_Polygonize(cells geometry);
CREATE OR REPLACE FUNCTION MAPI_Polygonize(cells geometry) 
  RETURNS TABLE (gid integer, geom geometry(polygon), area float8) AS
$BODY$
DECLARE
  startp timestamp;
BEGIN
	-- RAISE INFO 'MAPI_Polygonize .....';
	startp := clock_timestamp();
	RETURN QUERY 
		WITH
		q AS (
			SELECT ST_BuildArea(ST_Buffer(cells, 0.0001)) as ba
		),
		p AS (
			SELECT (ST_dump(ba)).geom AS pgeom
			FROM q
		)
		SELECT row_number() OVER (ORDER BY ST_Area(pgeom) DESC)::integer, pgeom, ST_Area(pgeom)
		FROM p;
	-- RAISE INFO '... DURATION=%    ', clock_timestamp() - startp;
END;
$BODY$ 
	LANGUAGE plpgsql VOLATILE;
COMMENT ON FUNCTION MAPI_Polygonize(geometry) IS
'Designed by Sylvain Piry, CBGP - INRA, 2013-2015
Aggregates contiguous cells from input collection in polygons.
Input parameters:
  1/ cells: geometry collection of polygons
Returns a set of rows: (geom geometry(polygon), area float8)
';




-- Designed by Sylvain Piry, CBGP - INRA, 2015-2016
-- Delineates signifcant polygons by merging contiguous significant cells.
-- Significance is computed given a alpha between 0.0 and 0.5 (typically 0.05 for 5%).
-- The "M_raw" method is a simple thresholding of p-values without False Discovery Rate processing.
-- Method "M_BY" (default) estimate the False Discovery Rate using Benjamini-Yekutieli (2001) algorithm.
-- Cells marked as significant are aggregated together in polygons.
-- These polygons are stored in two tables, one for the upper tail the other for the lower tail.
-- The table is named upon results table name and ends in _tails.
-- The table contains an unique id, the alpha value, the relation group name, the tail 
--     (lower or upper), the geometry of the polygon and its area.
-- If run with different alphas, polygons are added to the table with the alpha value used.
-- Input Parameters:
--   1/ The name of MAPI results table (without schema).
--   2/ The name of the schema.
--   3/ Name of the algorithm among: 
--       M_BY: Benjamini-Yekutieli FDR (default), 
--       M_raw: simple thresholding without FDR.
--   4/ alpha, default 0.05 (5%).
--   5/ Append - or not - to existing tables. If the table does not exists is created.
-- Returns the total number of polygons in lower and upper tail.
DROP FUNCTION IF EXISTS MAPI_Significance(results_table text, my_schema_name text, my_method varchar(250), my_alpha float8, append boolean);
CREATE OR REPLACE FUNCTION MAPI_Significance(results_table_name text, my_schema_name text, my_method varchar(250) DEFAULT 'M_BY', my_alpha float8 DEFAULT 0.05, append boolean DEFAULT false)
  RETURNS integer AS
$BODY$
DECLARE
	start1 timestamp;
	my_results_table text;
	srid_grid integer;
	final_tails text;
	groups_to_do text[];
	my_group text;
	resu_lower integer;
	resu_upper integer;
	resu integer;
BEGIN
	start1 := clock_timestamp();
	RAISE INFO 'DELINEATES SIGNIFICANT LOWER AND UPPER TAILS USING "%" ALGORITHM FOR ALPHA=%, ...', my_method, my_alpha;
	
	-- checkings
	IF (NOT my_alpha BETWEEN 0 AND 0.5) THEN RAISE 'Threshold range must be between 0.0 AND 0.5 (typically 0.05 for 5%%).'; END IF;
	
	-- full names
	my_results_table := '"'||coalesce(my_schema_name||'"."','')||results_table_name||'"';
	final_tails := '"'||coalesce(my_schema_name||'"."','')||results_table_name||'_tails"';
	
	-- get srid
	EXECUTE 'SELECT ST_SRID(geom) FROM '||my_results_table||' LIMIT 1;' 
		INTO srid_grid;
	
	-- drop previous tables, if any, and if not appending
	IF NOT append THEN
		EXECUTE 'DROP TABLE IF EXISTS '||final_tails||' CASCADE;';
	END IF;
	-- build new tables if needed
	EXECUTE 'CREATE TABLE IF NOT EXISTS '||final_tails||' (
				id serial PRIMARY KEY, 
				relation_group text, 
				alpha float8 NOT NULL,
				method varchar(250) NOT NULL,
				tail char(5) NOT NULL,
				geom geometry(polygon, '||srid_grid||'),
				area float8
			);';
			
	-- clean previous results (if any)
	EXECUTE 'DELETE FROM '||final_tails||' WHERE alpha = $1 AND method LIKE $2;' 
		USING my_alpha, my_method;
	
	IF (my_method LIKE 'M_BY') THEN
		-- processes lower tail
		EXECUTE 'WITH
				d(g, id, u, geom) AS (
					SELECT relation_group, id, pvalue, geom
					FROM '||my_results_table||'
					ORDER BY relation_group, pvalue
				),
				n(g, n) AS (
					SELECT g, count(id)
					FROM d
					GROUP BY g
				),
				q0(g, n, i) AS (
					SELECT g, n, generate_series(1,n)::float8
					FROM n
				),
				q(g, n, q) AS (
					SELECT g, n::float8, sum(1.0/i)
					FROM q0
					GROUP BY g, n
				),
				r(g, id, geom, u, n, q, r) AS (
					SELECT d.g, d.id, d.geom, d.u, q.n, q.q, 
							q.n * cume_dist() OVER (PARTITION BY d.g ORDER BY u)
					FROM d
					JOIN q ON d.g = q.g
					ORDER BY d.g, d.u
				),
				q2(g, cc) AS (
					SELECT g, ST_Collect(geom)
					FROM r
					WHERE u <= (r/(n*q))*$3
					GROUP BY g
				),
				q3(g, pol) AS (
					SELECT g, MAPI_Polygonize(cc)
					FROM q2
				),
				i AS (
					INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
						SELECT q3.g, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
						FROM q3
						RETURNING 1
				)
				SELECT count(*)
					FROM i;'
			USING my_method, 'lower', my_alpha
			INTO resu_lower;
		-- processes upper tail
		EXECUTE 'WITH
				d(g, id, u, geom) AS (
					SELECT relation_group, id, (1.0 - pvalue), geom
					FROM '||my_results_table||'
					ORDER BY relation_group, pvalue
				),
				n(g, n) AS (
					SELECT g, count(id)
					FROM d
					GROUP BY g
				),
				q0(g, n, i) AS (
					SELECT g, n, generate_series(1,n)::float8
					FROM n
				),
				q(g, n, q) AS (
					SELECT g, n::float8, sum(1.0/i)
					FROM q0
					GROUP BY g, n
				),
				r(g, id, geom, u, n, q, r) AS (
					SELECT d.g, d.id, d.geom, d.u, q.n, q.q, 
							q.n * cume_dist() OVER (PARTITION BY d.g ORDER BY u)
					FROM d
					JOIN q ON d.g = q.g
					ORDER BY d.g, d.u
				),
				q2(g, cc) AS (
					SELECT g, ST_Collect(geom)
					FROM r
					WHERE u <= (r/(n*q))*$3
					GROUP BY g
				),
				q3(g, pol) AS (
					SELECT g, MAPI_Polygonize(cc)
					FROM q2
				),
				i AS (
					INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
						SELECT q3.g, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
						FROM q3
						RETURNING 1
				)
				SELECT count(*)
					FROM i;'
			USING my_method, 'upper', my_alpha
			INTO resu_upper;
			
			
-- 		ELSIF (my_method LIKE 'M_BH') THEN
-- 			-- processes lower tail
-- 			EXECUTE 'WITH
-- 				d(g, id, u, geom) AS (
-- 					SELECT relation_group, id, pvalue, geom
-- 					FROM '||my_results_table||'
-- 					ORDER BY relation_group, pvalue
-- 				),
-- 				q(g, n, q) AS (
-- 					SELECT g, count(id), 1.0
-- 					FROM d
-- 					GROUP BY g
-- 				),
-- 				r(g, id, geom, u, n, q, r) AS (
-- 					SELECT d.g, d.id, d.geom, d.u, q.n, q.q, 
-- 							q.n * cume_dist() OVER (PARTITION BY d.g ORDER BY u)
-- 					FROM d
-- 					JOIN q ON d.g = q.g
-- 					ORDER BY d.g, d.u
-- 				),
-- 				q2(g, cc) AS (
-- 					SELECT g, ST_Collect(geom)
-- 					FROM r
-- 					WHERE u <= (r/(n*q))*$3
-- 					GROUP BY g
-- 				),
-- 				q3(g, pol) AS (
-- 					SELECT g, MAPI_Polygonize(cc)
-- 					FROM q2
-- 				),
-- 				i AS (
-- 					INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
-- 						SELECT q3.g, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
-- 						FROM q3
-- 						RETURNING 1
-- 				)
-- 				SELECT count(*)
-- 					FROM i;'
-- 			USING my_method, 'lower', my_alpha
-- 			INTO resu_lower;
-- 			-- processes upper tail
-- 			EXECUTE 'WITH
-- 				d(g, id, u, geom) AS (
-- 					SELECT relation_group, id, (1.0 - pvalue), geom
-- 					FROM '||my_results_table||'
-- 					ORDER BY relation_group, pvalue
-- 				),
-- 				q(g, n, q) AS (
-- 					SELECT g, count(id), 1.0
-- 					FROM d
-- 					GROUP BY g
-- 				),
-- 				r(g, id, geom, u, n, q, r) AS (
-- 					SELECT d.g, d.id, d.geom, d.u, q.n, q.q, 
-- 							q.n * cume_dist() OVER (PARTITION BY d.g ORDER BY u)
-- 					FROM d
-- 					JOIN q ON d.g = q.g
-- 					ORDER BY d.g, d.u
-- 				),
-- 				q2(g, cc) AS (
-- 					SELECT g, ST_Collect(geom)
-- 					FROM r
-- 					WHERE u <= (r/(n*q))*$3
-- 					GROUP BY g
-- 				),
-- 				q3(g, pol) AS (
-- 					SELECT g, MAPI_Polygonize(cc)
-- 					FROM q2
-- 				),
-- 				i AS (
-- 					INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
-- 						SELECT q3.g, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
-- 						FROM q3
-- 						RETURNING 1
-- 				)
-- 				SELECT count(*)
-- 					FROM i;'
-- 			USING my_method, 'upper', my_alpha
-- 			INTO resu_upper;
			
			
		ELSIF (my_method LIKE 'M_raw') THEN
			-- processes lower tail
			EXECUTE 'WITH
					q1 AS (
						SELECT relation_group, id, ( pvalue < $3 ) AS is_significant 
						FROM '||my_results_table||'
					),
					q2 AS (
						SELECT r.relation_group, ST_Collect(r.geom) AS cc
						FROM q1
						JOIN '||my_results_table||' AS r ON q1.id = r.id
						WHERE is_significant IS TRUE
						GROUP BY r.relation_group
					),
					q3 AS (
						SELECT q2.relation_group, MAPI_Polygonize(q2.cc) AS pol
						FROM q2
					),
					i AS (
						INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
							SELECT q3.relation_group, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
							FROM q3
							RETURNING 1
					)
					SELECT count(*)
						FROM i;'
				USING my_method, 'lower', my_alpha
				INTO resu_lower;
			
			-- processes upper tail
			EXECUTE 'WITH
					q1 AS (
						SELECT relation_group, id, ( pvalue > (1.0 - $3) ) AS is_significant 
						FROM '||my_results_table||'
					),
					q2 AS (
						SELECT r.relation_group, ST_Collect(r.geom) AS cc
						FROM q1
						JOIN '||my_results_table||' AS r ON q1.id = r.id
						WHERE is_significant IS TRUE
						GROUP BY r.relation_group
					),
					q3 AS (
						SELECT q2.relation_group, MAPI_Polygonize(q2.cc) AS pol
						FROM q2
					),
					i AS (
						INSERT INTO '||final_tails||' (relation_group, method, tail, alpha, geom, area)
							SELECT q3.relation_group, $1, $2, $3, (q3.pol).geom, ST_Area((q3.pol).geom)
							FROM q3
							RETURNING 1
					)
					SELECT count(*)
						FROM i;'
				USING my_method, 'upper', my_alpha
				INTO resu_upper;
		ELSE
			RAISE 'Unknown algorithm: "%".', my_method;
		END IF;
	
	RAISE INFO 'counts: lower=% upper=%... DURATION=%     ', coalesce(resu_lower, 0), coalesce(resu_upper, 0), clock_timestamp() - start1;
	resu := coalesce(resu_lower, 0) + coalesce(resu_upper, 0);
	RETURN resu;
END;
$BODY$
  LANGUAGE plpgsql VOLATILE;
COMMENT ON FUNCTION MAPI_Significance(results_table text, my_schema_name text, my_method varchar(250), my_alpha float8, append boolean) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2015-2016
Delineates signifcant polygons by merging contiguous significant cells.
Significance is computed given a alpha between 0.0 and 0.5 (typically 0.05 for 5%).
The "M_raw" method is a simple thresholding of p-values without False Discovery Rate processing.
Method "M_BY" (default) estimate the False Discovery Rate using Benjamini-Yekutieli algorithm.
Cells marked as significant are aggregated together in polygons.
These polygons are stored in two tables, one for the upper tail the other for the lower tail.
The table is named upon results table name and ends in _tails.
The table contains an unique id, the alpha value, the relation group name, the tail 
    (lower or upper), the geometry of the polygon and its area.
If run with different alphas, polygons are added to the table with the alpha value used.
Input Parameters:
  1/ The name of MAPI results table (without schema).
  2/ The name of the schema.
  3/ Name of the algorithm among: 
      M_BY: Benjamini-Yekutieli FDR (default), 
      M_raw: simple thresholding without FDR.
  4/ alpha, default 0.05 (5%).
  5/ Append - or not - to existing tables. If the table does not exists is created.
Returns the total number of polygons in lower and upper tail.
';


-- Designed by Sylvain Piry, CBGP - INRA, 2016
-- Computes the sum of squared differences between an array of values and a mean.
-- Parameters:
--   1/ a, the mean, float8.
--   2/ vv, the array of values, float8[].
-- Returns the sum of squares of differences.
DROP FUNCTION IF EXISTS MAPI_sd2(a float8, vv float8[]);
CREATE OR REPLACE FUNCTION MAPI_sd2(a float8, vv float8[])
	RETURNS float8 AS
$BODY$
DECLARE
	s float8;
BEGIN
	IF (array_length(vv, 1) = 1) THEN
		s := (a - vv[1])^2;
	ELSE 
		WITH q(v) AS (SELECT unnest(vv))
			SELECT sum( (a - v)*(a - v) ) FROM q WHERE v IS NOT NULL
		INTO s;
	END IF;
	return s;
END;
$BODY$ LANGUAGE plpgsql IMMUTABLE;

COMMENT ON FUNCTION MAPI_sd2(a float8, vv float8[]) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2016
Computes the sum of squared differences between an array of values and a mean.
Parameters:
  1/ a, the mean, float8.
  2/ vv, the array of values, float8[].
Returns the sum of squares of differences.
';


-- Designed by Sylvain Piry, CBGP - INRA, 2013-2016
-- Runs a full MAPI analysis provided views (or tables) containing points localizations and relationships between points and a template grid.
-- Parameters:
-- 1/ points_view_name: name of the table or view containing sampling points. Mandatory fields: point_name text, point_geom geometry(point), point_group text. The geometry must be projected (not lat/lon).
-- 2/ relations_view_name: name of the table or view containing the distance (or similarity) values between pairs of points. Mandatory fields: point_1_name text, point_2_name text, relation_value float8.
-- 3/ template_grid_name: name of the table or view containing the grid. Mandatory fields: gid integer primary key, geom geometry(polygon).
-- 4/ results_table_name: the name of the table that will be built with results (in schema "my_schema_name"). Warning! If the table already exists it is erased.
-- 5/ my_schema_name: the name of the schema containing the tables. Defaults to "public"
-- 6/ eccentricity: the eccentricity of the ellipses. Defaults to 0.975.
-- 7/ err_radius: minimal circle around points. Defaults to 10 projection units.
-- 8/ number of permutations. Defaults to 0 (no permutations).
-- 9/ also analyse relations between different groups. Boolean. Defaults to false.
-- 10/ minimal distance between two samples. Float. Defaults to NULL (unused).
-- 11/ maximal distance between two samples. Float. Defaults to NULL (unused).
DROP FUNCTION IF EXISTS MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8, options text);
CREATE OR REPLACE FUNCTION MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text DEFAULT NULL, eccentricity float8 DEFAULT 0.975, err_radius float8 DEFAULT 10, nb_permutations integer DEFAULT 1000, inter_groups boolean DEFAULT false, d_min float8 DEFAULT NULL, d_max float8 DEFAULT NULL, options text DEFAULT NULL)
  RETURNS integer AS
$BODY$
DECLARE
	clutch_envelope geometry;
	clutch_size int := 2048;
	clutches smallint[];
	current_group text;
	ellipses text;
	final text;
	final_comment text;
	final_perm text;
	ggids integer[];
	grid text;
	grpid1 int;
	grpid2 int;
	groups2 text;
	iloop int;
	individuals text;
	ind_groups text;
	ind_loc text;
	ind_loc_perm text;
	jloop int;
	locations text;
	locpairs_1gg text;
	max_id_loc_10 bigint;
	max_id_ind_10 bigint;
	max_id_grp_10 integer;
	my_ggid integer;
	myclutch smallint;
	nb INT;
	nb_cells int;
	nb_clutches integer;
	perm_max float8;
	perm_min float8;
	relations text;
	restab text;
	restab_perm text;
	run_id int;
	spatjoin text;
	srid_grid int;
	start0 timestamp;
	start1 timestamp;
	start2 timestamp;
	start3 timestamp;
	startx timestamp;
	weighted_ellipses text;
	weighted_ellipses_perm text;
BEGIN
	start0 := clock_timestamp();
	final_comment := format(E'MAPI_RunOnGrid\nStart: %s\n', start0);
	
	RAISE INFO 'START: ..... % ..... ------------------------------------------------------------------------------', clock_timestamp();
	RAISE INFO E'INPUT PARAMETERS:\n                points_view_name=%\n                relations_view_name=%\n                template_grid_name=%\n                results_table_name=%\n                my_schema_name=%\n                eccentricity=%\n                err_radius=%\n                nb_permutations=%\n                inter_groups=%\n                d_min=%\n                d_max=%\n                options=%\n', points_view_name, relations_view_name, template_grid_name, results_table_name, my_schema_name, eccentricity, err_radius, nb_permutations, inter_groups, d_min, d_max, options;
	
	final_comment := final_comment||E'INPUT PARAMETERS:\n';
	final_comment := final_comment||format(E'  points_view_name=%s\n', points_view_name);
	final_comment := final_comment||format(E'  relations_view_name=%s\n', relations_view_name);
	final_comment := final_comment||format(E'  template_grid_name=%s\n', template_grid_name);
	final_comment := final_comment||format(E'  results_table_name=%s\n', results_table_name);
	final_comment := final_comment||format(E'  my_schema_name=%s\n', coalesce(my_schema_name, 'NULL'));
	final_comment := final_comment||format(E'  eccentricity=%s\n', eccentricity);
	final_comment := final_comment||format(E'  err_radius=%s\n', err_radius);
	final_comment := final_comment||format(E'  nb_permutations=%s\n', nb_permutations);
	final_comment := final_comment||format(E'  inter_groups=%s\n', coalesce(inter_groups::text, 'NULL'));
	final_comment := final_comment||format(E'  d_min=%s\n', coalesce(d_min::text, 'NULL'));
	final_comment := final_comment||format(E'  d_max=%s\n', coalesce(d_max::text, 'NULL'));
	final_comment := final_comment||format(E'  options=%s\n', coalesce(options, 'NULL'));
	
	-- input values checkings
	IF (nb_permutations>0 AND nb_permutations<100) THEN RAISE WARNING 'Number of permutations should be at least >100 (or zero).'; END IF;
	-- get grid SRID
	EXECUTE 'SELECT ST_SRID(geom) FROM "'||coalesce(my_schema_name||'"."','')||template_grid_name||'" LIMIT 1' INTO srid_grid;
	-- and check it
	IF (get_proj4_from_srid(srid_grid) LIKE '%+proj=longlat%') THEN RAISE 'Points must be in cartesian coordinates.'; END IF;

	-- sequence for unique tables names
	IF NOT EXISTS (SELECT * FROM information_schema.sequences WHERE sequence_schema = 'public' AND sequence_name = 'mapi_seq') THEN
		CREATE SEQUENCE public.mapi_seq INCREMENT 1 MINVALUE 1 MAXVALUE 9223372036854775807 START 1 CACHE 1;
	END IF;
	run_id := nextval('mapi_seq');
		
	-- PREPARE TABLES
	individuals := 'tmp_mapi_individuals_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||individuals||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||individuals||' (
				idind serial PRIMARY KEY, 
				point_name text UNIQUE, 
				point_geom geometry(point, '||srid_grid||') NOT NULL, 
				point_group text NOT NULL, 
				grpid integer, 
				locid integer
			) ON COMMIT DROP;';
	
	ind_groups := 'tmp_mapi_groups_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||ind_groups||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||ind_groups||' (
				idgrp serial PRIMARY KEY, 
				group_name text UNIQUE
			) ON COMMIT DROP;';
	
	groups2 := 'tmp_mapi_group_pairs_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||groups2||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||groups2||' (
				idgg integer, 
				grpid1 integer, 
				grpid2 integer, 
				gg_name text
			) ON COMMIT DROP;';
	EXECUTE 'CREATE INDEX '||groups2||'_idgg_idx ON '||groups2||' USING btree(idgg);';
	EXECUTE 'CREATE INDEX '||groups2||'_idg2_idx ON '||groups2||' USING btree(grpid1, grpid2);';
	
	relations := 'tmp_mapi_relations_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||relations||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||relations||' (
				idrel bigint PRIMARY KEY, 
				iid1 integer, 
				iid2 integer, 
				ggid integer, 
				relation_value float8
			) ON COMMIT DROP;';
	
	locations := 'tmp_mapi_locations_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||locations||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||locations||' (
				idloc serial PRIMARY KEY, 
				ptgeom geometry(point, '||srid_grid||') NOT NULL
			) ON COMMIT DROP;';
	
	ellipses := 'tmp_mapi_unique_ellipses_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||ellipses||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||ellipses||' (
				idbiloc bigint PRIMARY KEY, 
				ellipse geometry(polygon), 
				distance float8, 
				inverse_area float8
			) ON COMMIT DROP;';

	grid := 'tmp_mapi_grid_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||grid||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||grid||' (
				gid integer PRIMARY KEY, 
				geom geometry(polygon, '||srid_grid||'), 
				ctr_geom geometry(point, '||srid_grid||'), 
				clutch smallint default 1::smallint
			) ON COMMIT DROP;';
	
	spatjoin := 'tmp_mapi_spatial_join_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||spatjoin||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||spatjoin||' (
				idsj bigserial PRIMARY KEY,
				bilocid bigint, 
				clutch smallint, 
				gid integer
			) ON COMMIT DROP;';
	
	ind_loc := 'tmp_mapi_ind_loc_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||ind_loc||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||ind_loc||' (
				iid integer, 
				locid integer
			) ON COMMIT DROP;';

	weighted_ellipses := 'tmp_mapi_weighted_relations_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||weighted_ellipses||' CASCADE;';
	EXECUTE 'CREATE TEMPORARY TABLE '||weighted_ellipses||' (
				bilocid bigint PRIMARY KEY, 
				ggid integer, 
				avg_value float8, 
				weight float8, 
				nb int,
				arr_values float8[]
			) ON COMMIT DROP;';
	
	restab := 'tmp_mapi_results_'||run_id;
	EXECUTE 'DROP TABLE IF EXISTS '||restab;
	EXECUTE 'CREATE TEMPORARY TABLE '||restab||' (
				id serial PRIMARY KEY, 
				gid integer, 
				ggid integer, 
				wmean_value float8, 
				wstddev_value float8,
				nb_rel integer, 
				sum_weights float8, 
				permuted_values float8[], 
				pvalue float8
			) ON COMMIT DROP';
	
	-- PREPARE TABLES FOR PERMUTATIONS
	IF (nb_permutations > 0) THEN
		ind_loc_perm := 'tmp_mapi_ind_loc_perm_'||run_id;
		EXECUTE 'DROP TABLE IF EXISTS '||ind_loc_perm||' CASCADE;';
		EXECUTE 'CREATE TEMPORARY TABLE '||ind_loc_perm||' (
					iid integer, 
					locid integer
				) ON COMMIT DROP;';
		
		weighted_ellipses_perm := 'tmp_mapi_weighted_relations_perm_'||run_id;
		EXECUTE 'DROP TABLE IF EXISTS '||weighted_ellipses_perm||' CASCADE;';
		EXECUTE 'CREATE TEMPORARY TABLE '||weighted_ellipses_perm||' (
					bilocid bigint PRIMARY KEY, 
					ggid integer, 
					avg_value float8, 
					weight float8, 
					nb int
				) ON COMMIT DROP;';
		
		restab_perm := 'tmp_mapi_results_perm_'||run_id;
		EXECUTE 'DROP TABLE IF EXISTS '||restab_perm||';';
		EXECUTE 'CREATE TEMPORARY TABLE '||restab_perm||' (
					id serial PRIMARY KEY, 
					gid integer, 
					ggid integer, 
					wmean_value float8
				) ON COMMIT DROP';
		
		final_perm := 'tmp_mapi_final_perm_'||run_id;
		EXECUTE 'DROP TABLE IF EXISTS '||final_perm||';';
		EXECUTE 'CREATE TEMPORARY TABLE '||final_perm||' (
					gid integer, 
					ggid integer, 
					iperm integer, 
					wmean_value float8
				) ON COMMIT DROP';
	END IF;
	
	-- final results table
	final := '"'||coalesce(my_schema_name||'"."','')||results_table_name||'"';
	EXECUTE 'DROP TABLE IF EXISTS '||final||' CASCADE';
	EXECUTE 'CREATE TABLE '||final||' (
				id serial PRIMARY KEY, 
				cell_gid integer, 
				relation_group text, 
				averaged_value float8, 
				weighted_stddev float8,
				sum_of_weights float8, 
				sum_of_weights_centile smallint, 
				number_of_ellipses integer, 
				permuted_values_str text, 
				pvalue float8, 
				geom geometry(polygon,'||srid_grid||')
			);';
			
	
	
	-- copy grid table
	start1 := clock_timestamp();
	RAISE INFO 'FILLS GRID TABLE...';
	EXECUTE 'WITH
			g(gid, geom) AS (
				SELECT min(gid), geom
				FROM "'||coalesce(my_schema_name||'"."','')||template_grid_name||'"
				GROUP BY geom
			)
			INSERT INTO '||grid||' (gid, geom, ctr_geom) 
				SELECT gid, geom, ST_Centroid(geom)
				FROM g;';
	EXECUTE 'SELECT count(*) FROM '||grid||';' INTO nb_cells;
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb_cells, clock_timestamp() - start1, clock_timestamp() - start0;
	
	-- computes clutches of cells when it worth it in order to avoid excessive memory use
	IF (nb_cells > 1.5*clutch_size) THEN
		start1 := clock_timestamp();
		RAISE INFO 'ASSIGNS CLUTCH # TO CELLS...';
		nb_clutches := nb_cells / clutch_size;
		EXECUTE 'WITH q(i,q) AS (
						SELECT gid, ntile($1) OVER (ORDER BY ST_X(ctr_geom), ST_Y(ctr_geom))::smallint 
						FROM '||grid||'
					)
					UPDATE '||grid||' AS g SET clutch = q.q FROM q WHERE g.gid = q.i;' 
				USING nb_clutches;
		EXECUTE 'SELECT array_agg(DISTINCT clutch ORDER BY clutch) FROM '||grid||';' INTO clutches;
		nb_clutches := array_length(clutches, 1);
		RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb_clutches, clock_timestamp() - start1, clock_timestamp() - start0;
	ELSE
		nb_clutches := 1;
		clutches := ARRAY[1::smallint];
	END IF;
	
	-- spatial index
	start1 := clock_timestamp();
	RAISE INFO 'BUILDS SPATIAL INDEX ON GRID TABLE...';
	EXECUTE 'CREATE INDEX '||grid||'_gist ON '||grid||' USING gist(geom);';
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb_cells, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- fills individuals table
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY INDIVIDUALS TABLE...';
	EXECUTE 'INSERT INTO '||individuals||' (point_name, point_geom, point_group) 
			SELECT point_name::text, ST_Transform(point_geom, '||srid_grid||'), coalesce(point_group::text, '''') 
			FROM "'||coalesce(my_schema_name||'"."','')||points_view_name||'"
			WHERE point_geom IS NOT NULL AND point_name IS NOT NULL;';
	EXECUTE 'SELECT (10^(char_length(max(idind)::text)))::bigint FROM '||individuals||';' INTO max_id_ind_10;
	EXECUTE 'SELECT count(*) FROM '||individuals||';' INTO nb;
	RAISE INFO 'count=%  ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- if option "pointsToCentroid" then move individual locations to their cell's centroid
	IF (options IS NOT NULL AND options ~* 'pointsToCentroid') THEN 
		start1 := clock_timestamp();
		RAISE INFO 'MOVE INDIVIDUALS LOCATIONS TO CELLS'' CENTROIDS...';
		EXECUTE 'WITH
				pc(idind, ctr) AS (
					SELECT i.idind, ST_Centroid(g.geom)
					FROM '||individuals||' AS i
					JOIN '||grid||' AS g ON ST_Intersects(i.point_geom, g.geom)
				)
				UPDATE '||individuals||' AS i SET point_geom = pc.ctr
					FROM pc
					WHERE i.idind = pc.idind';
		EXECUTE 'SELECT count(DISTINCT point_geom) FROM '||individuals||';' INTO nb;
		RAISE INFO 'count=%  ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	END IF;
	
	-- create groups
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY GROUPS TABLE AND UPDATES INDIVIDUALS...';
	-- define groups id
	EXECUTE 'WITH g AS (
					SELECT DISTINCT point_group 
					FROM '||individuals||'
				),
				ig(i, g) AS (
					SELECT row_number() OVER (ORDER BY point_group), point_group 
					FROM g
				)
				INSERT INTO '||ind_groups||' (idgrp, group_name)
					SELECT i, g 
					FROM ig;';
	EXECUTE 'UPDATE '||individuals||' AS ind SET grpid = grp.idgrp
				FROM '||ind_groups||' AS grp 
				WHERE ind.point_group = grp.group_name;';
	EXECUTE 'SELECT (10^(char_length(max(idgrp)::text)))::bigint FROM '||ind_groups||';' INTO max_id_grp_10;
	EXECUTE 'SELECT count(*) FROM '||ind_groups||';' INTO nb;
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- creates pairs of groups
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY PAIRS OF GROUPS TABLE...';
	IF (inter_groups) THEN
		EXECUTE 'INSERT INTO '||groups2||' (idgg, grpid1, grpid2, gg_name)
					SELECT least(g1.idgrp, g2.idgrp) * $1 + greatest(g1.idgrp, g2.idgrp),
						least(g1.idgrp, g2.idgrp), greatest(g1.idgrp, g2.idgrp), 
						g1.group_name||'',''||g2.group_name
					FROM '||ind_groups||' AS g1
					JOIN '||ind_groups||' AS g2 ON (g1.group_name <= g2.group_name);'
				USING max_id_grp_10;
	ELSE
		EXECUTE 'INSERT INTO '||groups2||' (idgg, grpid1, grpid2, gg_name) 
					SELECT idgrp * $1 + idgrp, idgrp, idgrp, group_name
					FROM '||ind_groups||';'
				USING max_id_grp_10;
	END IF;
	EXECUTE 'SELECT array_agg(DISTINCT idgg ORDER BY idgg) FROM '||groups2||';' INTO ggids; 
	nb := array_length(ggids, 1);
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- creates relation values table
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY RELATIONS TABLE...';
	EXECUTE 'WITH q(iid1, iid2, idgg, rvalue) AS (
				SELECT p1.idind, p2.idind, gg.idgg, r.relation_value::float8 
					FROM "'||coalesce(my_schema_name||'"."','')||relations_view_name||'" AS r
					JOIN '||individuals||' AS p1 ON p1.point_name = r.point_1_name::text
					JOIN '||individuals||' AS p2 ON p2.point_name = r.point_2_name::text
					JOIN '||groups2||' AS gg ON (  (p1.grpid = gg.grpid1 AND p2.grpid = gg.grpid2) 
												OR (p1.grpid = gg.grpid2 AND p2.grpid = gg.grpid1) )
					WHERE r.point_1_name::text NOT LIKE r.point_2_name::text
					AND r.relation_value IS NOT NULL
			)
			INSERT INTO '||relations||' (idrel, iid1, iid2, ggid, relation_value) 
				SELECT DISTINCT least(iid1, iid2)::bigint * $1 + greatest(iid1, iid2)::bigint, least(iid1, iid2), greatest(iid1, iid2), idgg, rvalue FROM q;'
		USING max_id_ind_10;
	EXECUTE 'SELECT count(*) FROM '||relations||';' INTO nb;
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- create distinct locations table (unpermuted)
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY DISTINCT LOCATIONS TABLE...';
	EXECUTE 'INSERT INTO '||locations||' (ptgeom) 
				SELECT DISTINCT point_geom 
				FROM '||individuals||';';
	EXECUTE 'SELECT count(*) FROM '||locations||';' INTO nb;
	EXECUTE 'SELECT (10^(char_length(max(idloc)::text)))::bigint FROM '||locations||';' INTO max_id_loc_10;
	-- assign location id to individuals
	EXECUTE 'UPDATE '||individuals||' AS p SET locid = l.idloc 
				FROM '||locations||' AS l 
				WHERE p.point_geom = l.ptgeom;';
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	
	-- creates distinct ellipses (optimized)
	start1 := clock_timestamp();
	RAISE INFO 'FILLS TEMPORARY DISTINCT ELLIPSES TABLE...';
	FOREACH my_ggid IN ARRAY ggids LOOP
		EXECUTE 'SELECT DISTINCT gg_name, grpid1, grpid2 FROM '||groups2||' WHERE idgg=$1;' USING my_ggid INTO current_group, grpid1, grpid2;
		RAISE INFO '    PROCESSING GROUP "%"', current_group;
		EXECUTE 'WITH
				lg(idloc, grpid) AS (
					SELECT DISTINCT l.idloc, i.grpid
					FROM '||locations||' AS l
					JOIN '||individuals||' AS i ON l.idloc = i.locid
					WHERE grpid IN ($3, $4)
				),
				dlp(id, lid1, lid2) AS (
					SELECT DISTINCT least(lg1.idloc, lg2.idloc)::bigint * '||max_id_loc_10||' + greatest(lg1.idloc, lg2.idloc)::bigint,
							least(lg1.idloc, lg2.idloc), greatest(lg1.idloc, lg2.idloc)
					FROM lg AS lg1
					CROSS JOIN lg AS lg2
					JOIN '||groups2||' AS gg ON gg.idgg = least(lg1.grpid, lg2.grpid) * '||max_id_grp_10||' + greatest(lg1.grpid, lg2.grpid)
					LEFT JOIN '||ellipses||' AS e ON least(lg1.idloc, lg2.idloc)::bigint * '||max_id_loc_10||' + greatest(lg1.idloc, lg2.idloc)::bigint = e.idbiloc
					WHERE e.distance IS NULL
				),
				ell(id, e, d) AS (
					SELECT dlp.id, MAPI_MakeEllipse(l1.ptgeom, l2.ptgeom, $1, $2), ST_Distance(l1.ptgeom, l2.ptgeom)
					FROM dlp
					JOIN '||locations||' AS l1 ON dlp.lid1 = l1.idloc
					JOIN '||locations||' AS l2 ON dlp.lid2 = l2.idloc '||
		CASE WHEN d_min IS NOT NULL THEN ' AND ST_Distance(l1.ptgeom, l2.ptgeom) >= '||d_min||' ' ELSE ' ' END ||
		CASE WHEN d_max IS NOT NULL THEN ' AND ST_Distance(l1.ptgeom, l2.ptgeom) <= '||d_max||' ' ELSE ' ' END ||'
				)
				INSERT INTO '||ellipses||' (idbiloc, ellipse, distance, inverse_area)
					SELECT id, e, d, 1.0/ST_Area(e)
					FROM ell;'
			USING  eccentricity, err_radius, grpid1, grpid2;
	END LOOP;
	EXECUTE 'SELECT count(*) FROM '||ellipses||';' INTO nb;
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	start1 := clock_timestamp();
	RAISE INFO 'BUILDS SPATIAL INDEX ON DISTINCT ELLIPSES TABLE...';
	EXECUTE 'CREATE INDEX '||ellipses||'_gist ON '||ellipses||' USING gist(ellipse);';
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- spatial join between grid and ellipses (clutch by clutch if necessary)
	start1 := clock_timestamp();
	jloop := 0;
	IF (nb_clutches > 1) THEN -- use clutches
		RAISE INFO 'FILLS TEMPORARY SPATIAL INTERSECT CLUTCH BY CLUTCH...';
		FOREACH myclutch IN ARRAY clutches LOOP
			start2 := clock_timestamp();
			jloop := jloop + 1;
			RAISE INFO '        SPATIAL INTERSECT ELLIPSES / CELLS CLUTCH # %/% [LONG!] ..... start at % .....', jloop, nb_clutches, clock_timestamp();
			EXECUTE 'SELECT ST_Envelope(ST_Collect(geom)) 
						FROM '||grid||' 
						WHERE clutch=$1;' 
					USING myclutch 
					INTO clutch_envelope;
			EXECUTE 'INSERT INTO '||spatjoin||' (bilocid, clutch, gid) 
						SELECT el.idbiloc, $1, g.gid
						FROM '||ellipses||' AS el
						JOIN '||grid||' AS g ON (el.ellipse && $2 AND clutch = $1 AND ST_Intersects(el.ellipse, g.geom));'
				USING myclutch, clutch_envelope;
			RAISE INFO '        ... DURATION=%    ', clock_timestamp() - start2;
		END LOOP;
	ELSE -- no clutches, use a global join
		RAISE INFO 'FILLS TEMPORARY SPATIAL INTERSECT ELLIPSES / ALL CELLS [LONG!] ..... start at % .....', clock_timestamp();
		-- startx := clock_timestamp();
		EXECUTE 'INSERT INTO '||spatjoin||' (bilocid, clutch, gid) 
					SELECT el.idbiloc, 1::smallint, g.gid
					FROM '||ellipses||' AS el
					JOIN '||grid||' AS g ON ST_Intersects(el.ellipse, g.geom);';
		-- RAISE INFO 'DURATION #4:    %    ', clock_timestamp() - startx; -- 00:01:18.986096
		
	END IF;
	RAISE INFO '... DURATION=%    ELAPSED=%    ', clock_timestamp() - start1, clock_timestamp() - start0;
	RAISE INFO 'BUILDS INDEX ON INTERSECTS TABLE...';
	EXECUTE 'CREATE INDEX '||spatjoin||'_idx ON '||spatjoin||' USING btree(clutch, bilocid);';
	EXECUTE 'SELECT count(*) FROM '||spatjoin||';' INTO nb;
	RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
	
	
	-- LOOP ON GROUPS
	FOREACH my_ggid IN ARRAY ggids LOOP
		EXECUTE 'SELECT DISTINCT gg_name FROM '||groups2||' WHERE idgg=$1;' USING my_ggid INTO current_group;
		
		RAISE INFO '********************************************** COMPUTING GROUP "%" **********************************************', current_group;
		
		-- unpermuted links table
		start1 := clock_timestamp();
		RAISE INFO 'FILLS TEMPORARY UNPERMUTED INDIVIDUALS LOCATIONS TABLE FOR GROUP "%" .....', current_group;
		EXECUTE 'TRUNCATE '||ind_loc||';';
		EXECUTE 'WITH u(iid, locid) AS (
					SELECT p.idind, p.locid
					FROM '||individuals||' AS p
					JOIN '||groups2||' AS gg ON (p.grpid = gg.grpid1 OR p.grpid = gg.grpid2)
					WHERE gg.idgg = $1
					ORDER BY p.idind
				)
				INSERT INTO '||ind_loc||' (iid, locid) 
					SELECT u.iid, u.locid
					FROM u;' USING my_ggid;
		EXECUTE 'SELECT count(*) FROM '||ind_loc||';' INTO nb;
		RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
		

		-- computes average value and weight for each ellipse
		start1 := clock_timestamp();
		RAISE INFO 'MERGES IDENTICAL PAIRS OF LOCATIONS FOR GROUP "%" .....', current_group;
		EXECUTE 'TRUNCATE '||weighted_ellipses||';';
		EXECUTE 'INSERT INTO '||weighted_ellipses||' (bilocid, ggid, avg_value, weight, nb, arr_values)
					SELECT el.idbiloc, $1, avg(r.relation_value), sum(el.inverse_area), count(*)::int, array_agg(r.relation_value)
					FROM '||relations||' AS r
					JOIN '||ind_loc||' AS il1 ON il1.iid = r.iid1
					JOIN '||ind_loc||' AS il2 ON il2.iid = r.iid2
					JOIN '||ellipses||' AS el ON el.idbiloc = least(il1.locid, il2.locid)::bigint * $2 + greatest(il1.locid, il2.locid)::bigint
					WHERE r.ggid = $1
					GROUP BY el.idbiloc;' 
				USING my_ggid, max_id_loc_10;
		EXECUTE 'SELECT count(*) FROM '||weighted_ellipses||';' INTO nb;
		RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
		
		
		-- computes weighted mean value by cell and group
		start1 := clock_timestamp();
		RAISE INFO 'COMPUTES WEIGHTED MEANS IN GRID FOR GROUP "%" .....', current_group;
		EXECUTE 'TRUNCATE '||restab||';';
		FOREACH myclutch IN ARRAY clutches LOOP
			EXECUTE 'INSERT INTO '||restab||' (gid, ggid, wmean_value, sum_weights, nb_rel)
						SELECT sj.gid, $1, sum(w.avg_value * w.weight) / sum(w.weight), sum(w.weight), sum(w.nb)
						FROM '||spatjoin||' AS sj
						JOIN '||weighted_ellipses||' AS w ON sj.clutch = $2 AND sj.bilocid = w.bilocid AND w.ggid = $1
						GROUP BY sj.gid;' USING my_ggid, myclutch;
		END LOOP;
		RAISE INFO '        ... DURATION=%    ELAPSED=%    ', clock_timestamp() - start1, clock_timestamp() - start0;
		
		
		-- computes weighted deviation value by cell and group
		-- for each ellipse the weight is constant, so we compute altogether related values
		start1 := clock_timestamp();
		RAISE INFO 'COMPUTES WEIGHTED STANDARD DEVIATION IN GRID FOR GROUP "%" .....', current_group;
		FOREACH myclutch IN ARRAY clutches LOOP
			EXECUTE 'WITH 
					q(gid, sumd2) AS (
						SELECT sj.gid, sum( MAPI_sd2(r.wmean_value, w.arr_values) * (w.weight / w.nb) )
						FROM '||spatjoin||' AS sj
						JOIN '||weighted_ellipses||' AS w ON sj.clutch = $2 AND sj.bilocid = w.bilocid AND w.ggid = $1
						JOIN '||restab||' AS r ON sj.gid = r.gid AND r.ggid = $1
						GROUP BY sj.gid
					)
					UPDATE '||restab||' AS r SET wstddev_value = sqrt(q.sumd2 / sum_weights)
					FROM q
					WHERE r.gid = q.gid AND r.ggid = $1;' USING my_ggid, myclutch;
		END LOOP;
		EXECUTE 'SELECT count(*) FROM '||restab||';' INTO nb;
		RAISE INFO 'count=% ... DURATION=%    ELAPSED=%    ', nb, clock_timestamp() - start1, clock_timestamp() - start0;
		
		
		
		/***************************************************************************************************************************************************/
		IF nb_permutations > 0 THEN
			FOR iloop IN 1 .. nb_permutations LOOP
				start2 := clock_timestamp();
				RAISE INFO '        GROUP "%": START OF PERMUTATION  # %/% .....', current_group, iloop, nb_permutations;
				
				-- permuted links table
				start3 := clock_timestamp();
				-- RAISE INFO '                CREATES PERMUTED INDIVIDUALS LOCATIONS TABLE FOR GROUP "%" .....', current_group;
				EXECUTE 'TRUNCATE '||ind_loc_perm||';';
				EXECUTE 'WITH 
						i AS (
							SELECT DISTINCT p.idind, p.locid
							FROM '||individuals||' AS p
							JOIN '||groups2||' AS gg ON (p.grpid = gg.grpid1 OR p.grpid = gg.grpid2)
							WHERE gg.idgg = $1
						),
						u(l, iid) AS (
							SELECT row_number() OVER (ORDER BY i.idind), i.idind
							FROM i
						),
						p(l, locid) AS (
							SELECT row_number() OVER (ORDER BY random()), i.locid
							FROM i
						)
						INSERT INTO '||ind_loc_perm||' (iid, locid) 
							SELECT u.iid, p.locid
							FROM u
							JOIN p ON u.l = p.l;' USING my_ggid;
				-- EXECUTE 'SELECT count(*) FROM '||ind_loc_perm||';' INTO nb;
				-- RAISE INFO '                count=% ... DURATION=%     ELAPSED=%    ', nb, clock_timestamp() - start3, clock_timestamp() - start0;
					
				-- computes average value and weight for each ellipse
				start3 := clock_timestamp();
				-- RAISE INFO '                MERGES IDENTICAL PAIRS OF LOCATIONS FOR PERMUTED GROUP "%" .....', current_group;
				EXECUTE 'TRUNCATE '||weighted_ellipses_perm||';';
				EXECUTE 'INSERT INTO '||weighted_ellipses_perm||' 
								(bilocid, ggid, avg_value, weight, nb)
							SELECT el.idbiloc, $1, avg(r.relation_value), sum(el.inverse_area), count(*)::int
							FROM '||relations||' AS r
							JOIN '||ind_loc_perm||' AS il1 ON il1.iid = r.iid1
							JOIN '||ind_loc_perm||' AS il2 ON il2.iid = r.iid2
							JOIN '||ellipses||' AS el ON el.idbiloc = least(il1.locid, il2.locid)::bigint * $2 + greatest(il1.locid, il2.locid)::bigint
							WHERE r.ggid = $1
							GROUP BY el.idbiloc;' 
						USING my_ggid, max_id_loc_10;
				-- EXECUTE 'SELECT count(*) FROM '||weighted_ellipses_perm||';' INTO nb;
				-- RAISE INFO '                count=% ... DURATION=%     ELAPSED=%    ', nb, clock_timestamp() - start3, clock_timestamp() - start0;
					
				-- computes value by cell and group
				start3 := clock_timestamp();
				-- RAISE INFO '                COMPUTES WEIGHTED MEANS IN GRID FOR PERMUTED GROUP "%" .....', current_group;
				EXECUTE 'TRUNCATE '||restab_perm||';';
				FOREACH myclutch IN ARRAY clutches LOOP
					EXECUTE 'INSERT INTO '||restab_perm||' (gid, ggid, wmean_value)
							SELECT sj.gid, $1, sum(w.avg_value * w.weight) / sum(w.weight)
							FROM '||spatjoin||' AS sj
							JOIN '||weighted_ellipses_perm||' AS w ON sj.clutch = $2 AND sj.bilocid = w.bilocid AND w.ggid = $1
							GROUP BY sj.gid;' 
						USING my_ggid, myclutch;
				END LOOP;
				-- EXECUTE 'SELECT count(*) FROM '||restab||';' INTO nb;
				-- RAISE INFO '                count=% ... DURATION=%     ELAPSED=%    ', nb, clock_timestamp() - start3, clock_timestamp() - start0;
				
				
				-- store final results in final_perm
				start3 := clock_timestamp();
				-- RAISE INFO '                INSERTING IN PERMUTATIONS TABLE PERMUTED GROUP "%" ..... ', current_group;
				EXECUTE 'INSERT INTO '||final_perm||' (gid, ggid, iperm, wmean_value)
						SELECT gid, ggid, $1, wmean_value
						FROM '||restab_perm||';' 
						USING iloop;
				-- RAISE INFO '                ... DURATION=%     ELAPSED=%    ', clock_timestamp() - start3, clock_timestamp() - start0;
				
				RAISE INFO '        END OF PERMUTATION  # %/% ... DURATION=%     ELAPSED=%    ', iloop, nb_permutations, clock_timestamp() - start2, clock_timestamp() - start0;
				
			END LOOP; -- on 1..nb_permutations

			RAISE INFO '------------------------------------------------------------------------------------------------------------';
			
			start1 := clock_timestamp();
			RAISE INFO 'INSERTING PERMUTED DISTRIBUTIONS IN RESULTS ..... ';
			EXECUTE 'WITH
					q AS (
						SELECT gid, ggid, array_agg(wmean_value ORDER BY iperm) AS pv
						FROM '||final_perm||'
						GROUP BY gid, ggid
					)
					UPDATE '||restab||' AS r SET permuted_values = q.pv, pvalue = NULL
					FROM q
					WHERE r.gid = q.gid AND r.ggid = q.ggid;';
			RAISE INFO '... DURATION=%     ELAPSED=%    ', clock_timestamp() - start1, clock_timestamp() - start0;
		END IF; -- nb_permutations > 0
		/***************************************************************************************************************************************************/
		
		-- store final results for this group
		start1 := clock_timestamp();
		RAISE INFO 'STORING IN FINAL TABLE ..... start at % .....', clock_timestamp();
		EXECUTE 'WITH 
				z(zone_geom) AS (
					SELECT ST_ConvexHull(ST_Collect(p.point_geom))
					FROM '||individuals||' AS p
					JOIN '||groups2||' AS gg ON (p.grpid = gg.grpid1 OR p.grpid = gg.grpid2)
					WHERE gg.idgg = $1 
				),
				i AS (
					INSERT INTO '||final||' (cell_gid, relation_group, averaged_value, weighted_stddev, sum_of_weights, sum_of_weights_centile, number_of_ellipses, permuted_values_str, pvalue, geom)
					SELECT r.gid, gg.gg_name, r.wmean_value, r.wstddev_value, r.sum_weights, ntile(100) OVER (PARTITION BY gg.gg_name ORDER BY r.sum_weights), 
							r.nb_rel, array_to_string(permuted_values, '',''), MAPI_RankInDistribution(r.wmean_value, r.permuted_values), g.geom
					FROM '||restab||' AS r
					JOIN '||grid||' AS g ON r.gid = g.gid
					JOIN '||groups2||' AS gg ON r.ggid = gg.idgg
					JOIN z ON ST_Intersects(z.zone_geom, g.geom)
					RETURNING 1
				)
				SELECT count(*) 
				FROM i;'
			USING my_ggid
			INTO nb;
					
		RAISE INFO '... DURATION=%     ELAPSED=%    ', clock_timestamp() - start1, clock_timestamp() - start0;
		RAISE INFO '********************************************** END OF GROUP "%" **********************************************', current_group;
	END LOOP; -- loop on groups
	
	EXECUTE 'SELECT count(*) FROM '||final||';' INTO nb;
	RAISE INFO '------------------------------------------------------------------------------------------------------------';
	
	final_comment := final_comment || format(E'End: %s (Duration=%s)\n', clock_timestamp(), clock_timestamp() - start0);
	EXECUTE 'COMMENT ON TABLE '||final||' IS '||quote_nullable(final_comment)||';';
	RAISE INFO E'ENDED.\n%', final_comment;
	RETURN nb;
END;
$BODY$ LANGUAGE plpgsql VOLATILE;
COMMENT ON FUNCTION MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8, options text) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2013-2016
Runs a MAPI analysis provided views (or tables) containing points localizations and 
    relationships between points and a template grid.
Input parameters:
  1/ points_view_name: name of the table or view containing sampling points. 
      Mandatory fields: point_name text, point_geom geometry(point), point_group text. 
      The geometry must be projected (not lat/lon).
  2/ relations_view_name: name of the table or view containing the distance (or similarity) 
      values between pairs of points. 
      Mandatory fields: point_1_name text, point_2_name text, relation_value float8.
  3/ template_grid_name: name of the table or view containing the grid. 
      Mandatory fields: gid integer primary key, geom geometry(polygon).
  4/ results_table_name: the name of the table that will be built for storing results 
      in schema "my_schema_name". 
      Warning! If the table already exists it is erased.
  5/ my_schema_name: the name of the schema containing the tables. Defaults to "public"
  6/ eccentricity: the eccentricity of the ellipses. Defaults to 0.975.
  7/ err_radius: minimal circle around points. Defaults to 10 projection units.
  8/ number of permutations. Defaults to 0 (no permutations).
  9/ also analyse relations between different groups. Boolean. Defaults to false.
 10/ minimal distance between two samples. Float. Defaults to NULL (unused).
 11/ maximal distance between two samples. Float. Defaults to NULL (unused).
 12/ options. Text "pointsToCentroid". Defaults to NULL (no options activated).
Returns the number of rows in results table.
';


-- Designed by Sylvain Piry, CBGP - INRA, 2015-2006
-- Run a MAPI analysis without grid provided. The grid is computed on the basis of a beta parameter wich may vary from 0.25 (random sampling) to 0.5 (regular sampling) in order to estimate the resolution of tje grid.
-- In the end, lower- and upper- tails are delineated after computation of FDR by Benjamini-Yekutieli method.
-- Parameters:
-- 1/ points_view_name: name of the table or view containing sampling points. Mandatory fields: point_name text, point_geom geometry(point), point_group text. The geometry must be projected (not lat/lon).
-- 2/ relations_view_name: name of the table or view containing the distance (or similarity) values between pairs of points. Mandatory fields: point_1_name text, point_2_name text, relation_value float8.
-- 3/ beta: Value of the sampling parameter from 0.5 (regular sampling) to 0.25 (random sampling). Smaller values will lead to finer grids.
-- 4/ results_table_name: the name of the table that will be built with results (in schema "my_schema_name"). Warning! If the table already exists it is erased.
-- 5/ my_schema_name: the name of the schema containing the tables. Defaults to "public"
-- 6/ eccentricity: the eccentricity of the ellipses. Defaults to 0.975.
-- 7/ err_radius: minimal circle around points. Defaults to 10 projection units.
-- 8/ nb_permutations. Defaults to 0 (no permutations).
-- 9/ inter_groups: also analyse relations between different groups. Boolean. Defaults to false.
-- 10/ d_min: minimal distance between two samples. Float. Defaults to NULL (unused).
-- 11/ d_max: maximal distance between two samples. Float. Defaults to NULL (unused).
-- 12/ alpha for significant areas. Float. Defaults to 0.05.
DROP FUNCTION IF EXISTS MAPI_RunAuto(points_view_name text, relations_view_name text, beta float8, results_table_name text, my_schema_name text, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8, alpha float8);
CREATE OR REPLACE FUNCTION MAPI_RunAuto(points_view_name text, relations_view_name text, beta float8, results_table_name text, my_schema_name text DEFAULT NULL, eccentricity float8 DEFAULT 0.975, err_radius float8 DEFAULT 10, nb_permutations integer DEFAULT 0, inter_groups boolean DEFAULT false, d_min float8 DEFAULT NULL, d_max float8 DEFAULT NULL, alpha float8 DEFAULT 0.05)
  RETURNS integer AS
$BODY$
DECLARE 
	grid_name text;
	nbc integer;
	nb integer;
	nbs integer;
BEGIN 
	RAISE INFO E'INPUT PARAMETERS:\n                points_view_name=%\n                relations_view_name=%\n                beta%\n                results_table_name=%\n                my_schema_name=%\n                eccentricity=%\n                err_radius=%\n                nb_permutations=%\n                inter_groups=%\n                d_min=%\n                d_max=%\n                alpha=%\n', points_view_name, relations_view_name, beta, results_table_name, my_schema_name, eccentricity, err_radius, nb_permutations, inter_groups, d_min, d_max, alpha;
	
	grid_name := results_table_name||'_grid';
	-- MAPI_GridAuto(points_view_name text, grid_table_name text, my_schema_name text, beta float8 DEFAULT 0.25)
	SELECT MAPI_GridAuto(points_view_name, grid_name, my_schema_name, beta) INTO nbc;

	-- MAPI_RunOnGrid(points_view_name text, relations_view_name text, template_grid_name text, results_table_name text, my_schema_name text DEFAULT NULL, eccentricity float8 DEFAULT 0.975, err_radius float8 DEFAULT 10, nb_permutations integer DEFAULT 1000, inter_groups boolean DEFAULT false, d_min float8 DEFAULT NULL, d_max float8 DEFAULT NULL, options text DEFAULT NULL)
	SELECT MAPI_RunOnGrid(points_view_name, relations_view_name, grid_name, results_table_name, my_schema_name, eccentricity, err_radius, nb_permutations, inter_groups, d_min, d_max, NULL) INTO nb;
	
	-- Significance and False Discovery Rate control
	IF nb_permutations > 0 THEN
		SELECT MAPI_Significance(results_table_name, my_schema_name, 'M_BY', alpha, false) INTO nbs;    -- drop _tails table first -> clean!
		-- We process only one method for "_tails" table simplicity
		-- SELECT MAPI_Significance(results_table_name, my_schema_name, 'M_raw', alpha, true) INTO nbs; -- add to _tails table
	END IF;
	
	RETURN nb;
END;
$BODY$ 
	LANGUAGE plpgsql VOLATILE;
COMMENT ON FUNCTION MAPI_RunAuto(points_view_name text, relations_view_name text, beta float8, results_table_name text, my_schema_name text, eccentricity float8, err_radius float8, nb_permutations integer, inter_groups boolean, d_min float8, d_max float8, alpha float8) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2013-2016
Run a MAPI analysis without grid provided. 
The grid is computed on the basis of a beta parameter wich may vary 
    from 0.25 (random sampling) to 0.5 (regular sampling) in order 
    to estimate the resolution of tje grid.
In the end, lower- and upper- tails are delineated after computation 
    of FDR by Benjamini-Yekutieli method.
Input parameters:
  1/ points_view_name: name of the table or view containing sampling points. 
      Mandatory fields: point_name text, point_geom geometry(point), point_group text. 
      The geometry must be projected (not lat/lon).
  2/ relations_view_name: name of the table or view containing the distance 
      (or similarity) values between pairs of points. 
      Mandatory fields: point_1_name text, point_2_name text, relation_value float8.
  3/ beta: Value of the sampling parameter from 0.5 (regular sampling) 
      to 0.25 (random sampling). Smaller values will lead to finer grids.
  4/ results_table_name: the name of the table that will be built with results 
      in schema "my_schema_name". 
      Warning! If the table already exists it is erased.
  5/ my_schema_name: the name of the schema containing the tables. 
      Defaults to "public"
  6/ eccentricity: the eccentricity of the ellipses. 
      Defaults to 0.975.
  7/ err_radius: minimal circle around points. 
      Defaults to 10 projection units.
  8/ nb_permutations. Number of permutations for significant areas detection.
      Defaults to 0 (no permutations).
  9/ inter_groups: also analyse relations between different groups. Boolean. 
      Defaults to false.
 10/ d_min: minimal distance between two samples. Float. 
      Defaults to NULL (unused).
 11/ d_max: maximal distance between two samples. Float. 
      Defaults to NULL (unused).
 12/ alpha for significant areas. Float. 
      Defaults to 0.05.
Returns the number of rows in results table.
';

-- replace plpgsql version of this function
LOAD '$libdir/mapi.so';
CREATE OR REPLACE FUNCTION MAPI_sd2(a float8, vv float8[])
RETURNS float8
	AS '$libdir/mapi.so', 'mapi_c_sd2'
	LANGUAGE C STRICT;
COMMENT ON FUNCTION MAPI_sd2(a float8, vv float8[]) IS 
'Designed by Sylvain Piry, CBGP - INRA, 2016
Computes the sum of squared differences between an array of values and a mean.
Parameters:
  1/ a, the mean, float8.
  2/ vv, the array of values, float8[].
Returns the sum of squares of differences.
';

