#include "postgres.h"
#include "utils/lsyscache.h"
#include "catalog/pg_type.h"
#include "utils/array.h"
#include "fmgr.h"
#include <math.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/* prototypes */
Datum mapi_c_sd2(PG_FUNCTION_ARGS);

/**
 * Computes the sum of the squares of differences between a mean and array values
**/
PG_FUNCTION_INFO_V1(mapi_c_sd2);

Datum mapi_c_sd2(PG_FUNCTION_ARGS) {
    ArrayType	*input;
    Oid 	i_eltype, a_eltype;
    int16	i_typlen;
    bool	i_typbyval;
    char	i_typalign;
    bool	*nulls;
    Datum	*i_data;
    int 	i, n;
    float8 	my_sum=0, a=0, d;

    /* return null on null input */
    if (PG_ARGISNULL(0) || PG_ARGISNULL(1)){
            PG_RETURN_NULL();
    }
    /* get input args */
    input = PG_GETARG_ARRAYTYPE_P(1);    
    /* get input array element type */
    i_eltype = ARR_ELEMTYPE(input);
    get_typlenbyvalalign(i_eltype, &i_typlen, &i_typbyval, &i_typalign);
    /* validate input array data type */
    switch(i_eltype){
        case FLOAT8OID:
            break;
        default:
            elog(ERROR, "Invalid 'array' data type");
            break;
    }
    
    /* get a data type */
    a_eltype = get_fn_expr_argtype(fcinfo->flinfo, 0);
    /* validate the a data type */
    switch(a_eltype){
        case FLOAT8OID: a = PG_GETARG_FLOAT8(0); break;
        default:
                elog(ERROR, "Invalid 'mean' data type");
                break;
    }
    
    /* get src data */
    deconstruct_array(input, i_eltype, i_typlen, i_typbyval, i_typalign, &i_data, &nulls, &n);
   
    /* compute sum of squared differences values */
	if (n > 0) {
		my_sum = 0.0;
		for (i=0; i < n; i++)  {
			if (! nulls[i]) {
				d = (a - DatumGetFloat8(i_data[i]));
				my_sum += d*d;
			}
		}
		PG_RETURN_FLOAT8(my_sum);
	} else {
		PG_RETURN_NULL();
	}
}
